/*+
 * File:	gd_graph.c
 *
 * Purpose:	Provides implementations of graph routines
 *              defined in gd_graph.h
 *
 * Author:	Thomas R. Dean
 *
 * Revision History:
 *
 *  1.0 - January-February 2001
 *      - initial version
 -*/

#include "stdio.h"
#include "stdlib.h"

#include "gd_graph.h"

const char *gd_node_type_ascii[NUM_NODES + 1] = {

  /* gcc node names */
#define DEFTREECODE(SYM,STR,A,B) STR,
#include "tree.def"
#include "c-common.def"
#include "cp/cp-tree.def"
#undef DEFTREECODE

  /* datrix node names */
  "AsgNd",
    "Identifier",
      "Type",
	"BuiltInType",
	"AggrType",
	"EnumType",
	"TemplParmType",
	"GenericType",
	"GeneratedType",
	"AliasType",
	"ForwardType",
	"ArrayType",
	"PtrType",
	"RefType",
	"FctType",
      "Scope",
	"System",
	"ScopeGlb",
	"ScopeCompil",
	"ScopeFile",
	"Package",
	"NameSpace",
	"Block",
	"StaticBlock",
	"TryBlock",
	"CatchBlock",
	"FinallyBlock",
	"SynchronizedBlock",
      "Object",
	"FormalFctParam",
	"FormalCatchParam",
      "Function",
      "Label",
	"CaseLabel",
	"DefaultLabel",
      "Enumerator",
    "Expression",
      "EmptyExpr",
      "Literal",
      "SyntaxError",
      "NameRef",
      "Operator",
	"UnaryOperator",
	"BinaryOperator",
	"QuestionOperator",
      "InitList",
      "ActualFctParamList",
      "ActualTemplParamList",
      "CreationExpr",
      "DestroyExpr",
    "CntrlStmt",
      "Selection",
	"If",
	"Switch",
      "Jump",
	"Goto",
	"Return",
	"Break",
	"Continue",
      "Loop",
	"DoWhileLoop",
	"WhileDoLoop",
	"ForLoop",
    "SqlStmt",
    "Comment",
      "JavaDocComment",
    "ThrowSpec",
    "AsmBlock",
  "Arc",
    "ArcReference",
      "Instance",
      "Inheritance",
      "Implements",
      "ArcFriend",
      "ArcDefLoc",
      "RefersTo",
    "ArcSyntax",
      "ArcSon",
      "ArcTrue",
      "ArcFalse",
      "ArcOpd",
      "Import",
      "ArcCond",
      "ArcForInit",
      "ArcForIncr",
      "ArcInitVal",
      "ArcBitDim",
      "ArcArrayDim",

  /* end of types from DATRIX */

  /* deleted node */
  "node_delted",

  /* internal use nodes types */

  "node_none",
  /* end internal use node types */
  "num_nodes"
};

gd_graph allocate_gd_graph(unsigned long StringSize, unsigned long HashSize,
    			unsigned long GraphSize) {
  gd_graph result;
  unsigned int i;

  /* initial null result */
  result.stringBuffLen = 0;
  result.curStringPos = 0;
  result.stringBuffer = NULL;
  result.hashTableLen = 0;
  result.hashTable = NULL;
  result.nodeBuffLen = 0;
  result.curNodePos = 0;
  result.firstNode = -1;
  result.nodeBuffer = NULL;


  /* sanity checks  - just enough to make sure other routines don't crash */
  if (StringSize < 2) return result;
  if (HashSize < 2) return result;
  if (GraphSize < 2) return result;

  /* allocate the structure */
  result.stringBuffLen = StringSize;
  result.curStringPos = 0;
  result.stringBuffer = (char *) malloc(StringSize);
  if (result.stringBuffer == NULL) {
    /* memory allocation error return to initial value */
    result.stringBuffLen = 0;
    result.curStringPos = 0;
    return result;
  }

  result.hashTableLen = HashSize;
  result.hashTable = (long *) malloc(HashSize * sizeof(unsigned long));
  if (result.hashTable == NULL) {
    /* memory allocation error return to initial value */
    result.stringBuffLen = 0;
    result.curStringPos = 0;
    free(result.stringBuffer); /* free previously allocated memory */
    result.stringBuffer = NULL;
    result.hashTableLen = 0;
    return result;
  }
  /* initialize hash table */
  for(i = 0; i < result.hashTableLen; i++){
    result.hashTable[i] = -1; /* -1 means unallocated */
  }

  result.nodeBuffLen = GraphSize;
  result.curNodePos = 0;
  result.nodeBuffer = (gd_graph_node *) malloc(GraphSize * sizeof(gd_graph_node));
  if (result.nodeBuffer == NULL) {
    /* memory allocation error return to initial value */
    result.stringBuffLen = 0;
    result.curStringPos = 0;
    free(result.stringBuffer); /* free previously allocated memory */
    result.stringBuffer = NULL;
    result.hashTableLen = 0;
    free(result.hashTable); /* free previously allocated memory */
    result.hashTable = NULL;
    result.nodeBuffLen = 0;
    result.curNodePos = 0;
    return result;
  }
  return result;
}

void deallocate_gd_graph(gd_graph g) {
  free(g.stringBuffer);
  free(g.hashTable);
  free(g.nodeBuffer);
}

/***********************************************************
 * Binary File I/O
 * File format:
 *     4 bytes - magic ('gduf')
 *     4 bytes - version, first two bytes are major, second
 *       two bytes are minor
 *     4 bytes - length of character table, value = X
 *     X bytes - character table
 *     4 bytes - length of symbol table, value = Y
 *     Y words - symbol table
 *     4 bytes - length of node table, value = Z
 *     Z nodes - node table
 ***********************************************************/

static char gd_magic[4] = { 'g', 'd', 'u', 'f' };

/* returns 1 for success, 0 for failure */
int gd_graph_write_binary_file(const char * filename, gd_graph g) {

  unsigned short version[2];
  size_t numItems;
  FILE * f;

  if ((f = fopen(filename,"wb")) == NULL) {
    fprintf(stderr,"Could not open %s for binary write\n",filename);
    return 0;
  }

  if ((numItems = fwrite(gd_magic,sizeof(char), 4, f)) < 4) {
    fprintf(stderr,"Error writing magic to %s\n",filename);
    fclose(f);
    return 0;
  }

  version[0] = GRAPH_MAJOR;
  version[1] = GRAPH_MINOR;
  if ((numItems = fwrite(version,sizeof(unsigned short), 2,f)) < 2) {
    fprintf(stderr,"Error writing version to %s\n",filename);
    fclose(f);
    return 0;
  }

  /* write the string buffer */
  /* curString pos points to the first free byte in the buffer, which
   * is the length of the used part of the buffer */
  if ((numItems = fwrite(&g.curStringPos,sizeof(unsigned long),1,f)) < 1){
    fprintf(stderr,"Error writing string buffer length to %s\n",filename);
    fclose(f);
    return 0;
  }
  if ((numItems = fwrite(g.stringBuffer,sizeof(char),g.curStringPos,f)) < g.curStringPos){
    fprintf(stderr,"Error writing string buffer to %s\n",filename);
    fclose(f);
    return 0;
  }

  /* write the hash table */
  /* we have to write the whole hash table */
  if ((numItems = fwrite(&g.hashTableLen,sizeof(unsigned long),1,f)) < 1){
    fprintf(stderr,"Error writing hash table length to %s\n",filename);
    fclose(f);
    return 0;
  }
  if ((numItems = fwrite(g.hashTable,sizeof(long),g.hashTableLen,f)) < g.hashTableLen){
    fprintf(stderr,"Error writing hash table to %s\n",filename);
    fclose(f);
    return 0;
  }

  /* write first node */
  if ((numItems = fwrite(&g.firstNode,sizeof(node_ptr),1,f)) < 1){
    fprintf(stderr,"Error writing fist node to %s\n",filename);
    fclose(f);
    return 0;
  }
  /* write the node table */
  /* as with the string table, we only write what is used */
  if ((numItems = fwrite(&g.curNodePos,sizeof(unsigned long),1,f)) < 1){
    fprintf(stderr,"Error writing node buffer length to %s\n",filename);
    fclose(f);
    return 0;
  }
  if ((numItems = fwrite(g.nodeBuffer,sizeof(gd_graph_node),g.curNodePos,f)) < g.curNodePos){
    fprintf(stderr,"Error writing node buffer to %s\n",filename);
    fclose(f);
    return 0;
  }

  fclose(f);
  return 1;
}

/* similar to write, but in reverse.
 * also allocates more space depending on last three parameters */
gd_graph gd_graph_read_binary_file(const char * filename, float string_mult, 
				  float hash_mult, float node_mult) {

  unsigned short version[2];
  char magic_in[4];
  unsigned long tmp;
  size_t numItems;
  FILE * f;
  gd_graph result;
  int i;

  /* initial null result */
  result.stringBuffLen = 0;
  result.curStringPos = 0;
  result.stringBuffer = NULL;
  result.hashTableLen = 0;
  result.hashTable = NULL;
  result.nodeBuffLen = 0;
  result.curNodePos = 0;
  result.nodeBuffer = NULL;

  /* sanity check, cannot reduce size of structures */
  if (string_mult < 1.0) return result;
  if (hash_mult < 1.0) return result;
  if (node_mult < 1.0) return result;

  if ((f = fopen(filename,"rb")) == NULL) {
    fprintf(stderr,"Could not open %s for binary read\n",filename);
    return result;
  }

  if ((numItems = fread(magic_in,sizeof(char), 4,f)) < 4) {
    fprintf(stderr,"Error reading magic from %s\n",filename);
    fclose(f);
    return result;
  }

  /* check magic */
  for (i = 0; i < 4; i++) {
    if (magic_in[i] != gd_magic[i]) {
      fprintf(stderr,"%s: Wrong file format, magic does not match\n",filename);
      fclose(f);
      return result;
    }
  }
  

  if ((numItems = fread(version,sizeof(unsigned short), 2,f)) < 2) {
    fprintf(stderr,"Error reading version from %s\n",filename);
    fclose(f);
    return result;
  }

/* fprintf(stderr,"version %d,%d\n",version[0],version[1]); */

  if (version[0] != GRAPH_MAJOR){
    fprintf(stderr,"Major version %d in file %s does not match current version %d\n",
		  version[0],filename,GRAPH_MAJOR);
    fclose(f);
    return result;
  }
  /* minor upgrades are compatible, that is V2.3 will read V2.2 files */
  if (version[1] > GRAPH_MINOR){
    fprintf(stderr,"Minor version %d in file %s is newer than current version %d\n",
		  version[1],filename,GRAPH_MINOR);
    fclose(f);
    return result;
  }

  /* read the string buffer */
  /* curString pos points to the first free byte in the buffer, which
   * is the length of the used part of the buffer */
  if ((numItems = fread(&tmp,sizeof(unsigned long),1,f)) < 1){
    fprintf(stderr,"Error reading string buffer length from %s\n",filename);
    fclose(f);
    return result;
  }
  result.curStringPos = tmp;

/* fprintf(stderr,"numChars=%d\n",result.curStringPos); */

  /* allow us to increase size of string buffer */
  result.stringBuffLen = (unsigned long) (result.curStringPos * string_mult) + 1;
  result.stringBuffer = (char *)malloc(result.stringBuffLen);
  if (result.stringBuffer == NULL) {
    fprintf(stderr,"Could not allocate string buffer when reading %s\n",filename);
    fclose(f);
    result.curStringPos = 0;
    result.stringBuffLen = 0;
    return result;
  }

  /* read the buffer from the file.  the tail end of the buffer will be empty */
  if ((numItems = fread(result.stringBuffer,sizeof(char),result.curStringPos,f)) < result.curStringPos){
    fprintf(stderr,"Error reading string buffer from %s\n",filename);
    fclose(f);
    result.curStringPos = 0;
    result.stringBuffLen = 0;
    free(result.stringBuffer);
    result.stringBuffer = NULL;
    return result;
  }

  /* read the hash table */
  if ((numItems = fread(&tmp,sizeof(unsigned long),1,f)) < 1){
    fprintf(stderr,"Error reading hash table length from %s\n",filename);
    fclose(f);
    result.curStringPos = 0;
    result.stringBuffLen = 0;
    free(result.stringBuffer);
    result.stringBuffer = NULL;
    return result;
  }
  result.hashTableLen = tmp;

/* fprintf(stderr,"hashTable %d\n",result.hashTableLen); */

  result.hashTable = (long *)malloc(result.hashTableLen * sizeof(long));
  if (result.hashTable == NULL) {
    fprintf(stderr,"Could not allocate hash table reading from file %s\n",filename);
    fclose(f);
    result.curStringPos = 0;
    result.stringBuffLen = 0;
    free(result.stringBuffer);
    result.stringBuffer = NULL;
    result.hashTableLen = 0;
    return result;
  }

  if ((numItems = fread(result.hashTable,sizeof(long),result.hashTableLen,f)) < result.hashTableLen){
    fprintf(stderr,"Error reading hash table from %s\n",filename);
    fclose(f);
    result.curStringPos = 0;
    result.stringBuffLen = 0;
    free(result.stringBuffer);
    result.stringBuffer = NULL;
    result.hashTableLen = 0;
    free(result.hashTable);
    result.hashTable = NULL;
    return result;
  }

  if (hash_mult > 1.0) {
    /* TODO rehash */
    fprintf(stderr,"Hash Table Resize not implemented yet\n");
    fclose(f);
    result.curStringPos = 0;
    result.stringBuffLen = 0;
    free(result.stringBuffer);
    result.stringBuffer = NULL;
    result.hashTableLen = 0;
    free(result.hashTable);
    result.hashTable = NULL;
    return result;
  }

  /* read the first node */
  if ((numItems = fread(&(result.firstNode),sizeof(node_ptr),1,f)) < 1){
    fprintf(stderr,"Error reading node buffer length from %s\n",filename);
    fclose(f);
    result.curStringPos = 0;
    result.stringBuffLen = 0;
    free(result.stringBuffer);
    result.stringBuffer = NULL;
    result.hashTableLen = 0;
    free(result.hashTable);
    result.hashTable = NULL;
    return result;
  }
  /* read the node table */
  /* as with the string table, we can resize on input */
  if ((numItems = fread(&tmp,sizeof(unsigned long),1,f)) < 1){
    fprintf(stderr,"Error reading node buffer length from %s\n",filename);
    fclose(f);
    result.curStringPos = 0;
    result.stringBuffLen = 0;
    free(result.stringBuffer);
    result.stringBuffer = NULL;
    result.hashTableLen = 0;
    free(result.hashTable);
    result.hashTable = NULL;
    result.firstNode = -1;
    return result;
  }
  result.curNodePos = tmp;

/* fprintf(stderr,"curNodePos %d\n",result.curNodePos); */

  result.nodeBuffLen = (unsigned long) (result.curNodePos * node_mult) + 1;

  result.nodeBuffer = (gd_graph_node *) malloc(result.nodeBuffLen * sizeof(gd_graph_node));
  if (result.nodeBuffer == NULL) {
    fprintf(stderr,"Error allocating node buffer");
    fclose(f);
    result.curStringPos = 0;
    result.stringBuffLen = 0;
    free(result.stringBuffer);
    result.stringBuffer = NULL;
    result.hashTableLen = 0;
    free(result.hashTable);
    result.hashTable = NULL;
    result.curNodePos = 0;
    result.nodeBuffLen = 0;
    result.firstNode = -1;
    return result;
  }

  if ((numItems = fread(result.nodeBuffer,sizeof(gd_graph_node),result.curNodePos,f)) < result.curNodePos){
    fprintf(stderr,"Error reading node buffer from %s(%d,%ld)\n",filename,numItems, result.curNodePos);
    fclose(f);
    result.curStringPos = 0;
    result.stringBuffLen = 0;
    free(result.stringBuffer);
    result.stringBuffer = NULL;
    result.hashTableLen = 0;
    free(result.hashTable);
    result.hashTable = NULL;
    result.curNodePos = 0;
    result.nodeBuffLen = 0;
    free(result.nodeBuffer);
    result.nodeBuffer = NULL;
    result.firstNode = -1;
    return result;
  }

  fclose(f);
  return result;
}

/*********************************************
 * String Ops 
 *********************************************/

/* hash algorithms */
static unsigned long primary_hash(const char * inString,unsigned long len,
						unsigned long tableSize){
  /* special case for null string */
  if (len == 0) return 1;
  return (((inString[0] << 7) + (inString[len-1] << 4) + len) % tableSize);
}

static unsigned long secondary_hash(const char * inString,unsigned long len,
			      unsigned long prev, unsigned long tableSize){
  /* special case for null string */
  if (len == 0) return prev + 1;
  /* char value of string is not used in secondary */
  return ((prev + len) % tableSize);
}

/* do the acutal hash table manipulation */
/* returns the hash location of the string, or
 * the first empty slot, returns -1 if there is
 * no empty slot */

static long do_hash(gd_graph g, const char * string){
  unsigned long hashVal, firstHash;
  unsigned long sLength;

  sLength = strlen(string);
  hashVal = primary_hash(string,sLength, g.hashTableLen);

  if (g.hashTable[hashVal] != -1) {
    /* something in the primary hash slot */
    if (strcmp(&(g.stringBuffer[g.hashTable[hashVal]]),string) == 0){
      /* string found -- return hash location */
      return hashVal;
    }
    /* different string in the same hash slot */
    firstHash = hashVal;
    do {
      hashVal = secondary_hash(string,sLength,hashVal,g.hashTableLen);
      if (g.hashTable[hashVal] != -1) {
	if (strcmp(&(g.stringBuffer[g.hashTable[hashVal]]),string) == 0){
	  /* string found -- return hash location */
	  return hashVal;
	}
      }
    } while (g.hashTable[hashVal] != -1 && hashVal != firstHash);
    if (firstHash == hashVal) {
      /* cycled entire table without finding an empty slot */
      return -1;
    }
  }
  /* not found, return location of empty slot */
  return hashVal;
}

/* -1 means not found */
long gd_lookup_string(gd_graph *g, const char * string) {
  long hashVal = do_hash(*g,string);
  if (hashVal == -1) return -1;
  if (strcmp(&(g->stringBuffer[g->hashTable[hashVal]]),string) == 0){
    /* hash location has location of string, return string location */
    return g->hashTable[hashVal];
  }
  return -1;
}

/* same as lookup, but if we don't find it, add it to the table, -1 means an error */
long gd_add_string(gd_graph *g, const char * newstring){
  long hashVal = do_hash(*g,newstring);
  unsigned long stringLen;

  if (hashVal == -1) return -1; /* no space left in hash table */

  if (g->hashTable[hashVal] != -1){
    /* hash location has location of string, return string location */
    return g->hashTable[hashVal];
  }

  /* not found
   *   -- insert it
   *   -- hash value is an empty slot */ 

  /* check for space in stringBuffer */
  stringLen = strlen(newstring) + 1;
  if (g->stringBuffLen - g->curStringPos < stringLen) {
    fprintf(stderr,"out of string buffer space");
    return -1;
  }
  g->hashTable[hashVal] = g->curStringPos; /* store location of string contents */
  /* copy the string contents to the string buffer */
  strcpy(&(g->stringBuffer[g->curStringPos]),newstring);
  g->curStringPos += stringLen;
  return g->hashTable[hashVal];
}

int gd_grow_nodes(gd_graph * g, unsigned long grow_amt) {

  unsigned long newSize;
  gd_graph_node * newBuffer, *oldBuffer;

  newSize = g->nodeBuffLen + grow_amt;
  newBuffer = (gd_graph_node *) malloc(newSize * sizeof(gd_graph_node));
  if (newBuffer == NULL){
    return 0;
  }
  
  oldBuffer = g->nodeBuffer;
  memcpy(newBuffer,oldBuffer,g->nodeBuffLen * sizeof(gd_graph_node));
  g->nodeBuffer = newBuffer;
  g->nodeBuffLen = newSize;
  free(oldBuffer);
  return(1);
}

