/*+
 * Name:	aggregate.c
 *
 * Purpose:	Transforms gcc record_type and union_type to datrix aggr_type.
 *              -- transfomrs flds and fncs in gcc type to arcSon
 *              -- copies name from identifier node in decl node to datrix type node
 *              -- copies file and line info
 *
 * Author:	Thomas R. Dean
 *
 * Date:	March 7, 2001
 *
-*/

#include <stdio.h>
#include <stdlib.h>

#include "gd_graph.h"
#include "gd_utils.h"

char * progname;

void usage(){
  fprintf(stderr,"usage: %s infile outfile\n",progname);
  exit(1);
}

void unifyAggregates(gd_graph * g, node_ptr n, node_ptr *f, int i);
void do_aggregate(gd_graph *g,int i);
int isNamed(gd_graph * g,node_ptr funct_node, str_ptr name);

str_ptr base_ctor, comp_ctor;

int main(int argc, char * argv[]){
  gd_graph g;
  char * filename, *outfilename;
  int nodenum,i, count;

  progname = argv[0];

  if (argc != 3) {
    usage();
  }

  filename = argv[1];
  outfilename = argv[2];

  g = gd_graph_read_binary_file(filename,1.0,1.0,1.0);
  if (g.curStringPos == 0) {
    fprintf(stderr,"could not get graph\n");
    exit(1);
  }

  /* there may be more than one record_type node for
   * a given struct, class or union. Unify them */
  walk_ptrs(&g,unifyAggregates);

  /* prefetch names for extra constructors
   */
  base_ctor = gd_lookup_string(&g, "__base_ctor");
  comp_ctor = gd_lookup_string(&g, "__comp_ctor");

  /* convert aggregate nodes */
  for( i = 0; i < g.curNodePos; i++){
    if (g.nodeBuffer[i].common.type == gRECORD_TYPE ||
	g.nodeBuffer[i].common.type == gUNION_TYPE){
      if (g.nodeBuffer[i].type.unql == i){
	/* only convert primary aggregate nodes */
	do_aggregate(&g,i);
      }
    }
  }

  /* all done, write the file */
  if (gd_graph_write_binary_file(outfilename,g) == 0){
    fprintf(stderr,"problem writing file\n");
    exit(1);
  }

  return 0;
}

/*
 * unifyAggregates -- sometimes GCC duplicates a record_type
 * node for paramters to functions. These node should be unified
 * with the original declarations of the types. Essentially,
 * there should be an unql pointer chain that leads to the
 * orginal definition. A shortcut is to follow the name pointer
 * if it is defined (gives a type_decl), and then from there to the type.
 * If that gives us the same node we are at, then we are at the original definition.
 * If not, then we change the pointer to the original defintiion.
 *
 * If the name pointer is not defined, or the graph doesn't match what
 * we expect, follow the unql pointer to the end.
 *
 */

void unifyAggregates(gd_graph * g, node_ptr n, node_ptr *f, int i) {
  /* parms n, i unused */
  node_ptr name;
  node_ptr type;
  node_ptr unql, prevunql;

  if (*f == -1) return;

  if (*f < -1 || *f > g-> curNodePos) {
    fprintf(stderr,"wild pointer (%d) node %d, index %d\n",*f,n,i);
    exit(1);
  }

  if (g->nodeBuffer[*f].common.type == gRECORD_TYPE ||
      g->nodeBuffer[*f].common.type == gUNION_TYPE){
    /* first check name field */
    if ((name = g -> nodeBuffer[*f].type.name) != -1) {
      if (g->nodeBuffer[name].common.type == gTYPE_DECL){
	if ((type = g-> nodeBuffer[name].decl.type) != *f){
	  /* name ptr points to node whose type pointer does not point back to us!! */
	  /* therefore we are a copy, verify by checking that our unql pointer is not
	   * reflexive */
	  if (g-> nodeBuffer[*f].type.unql == *f){
	    fprintf(stderr,"name ptr says copied aggregate type, but uql says not (%d)!!\n",*f);
	    exit(1);
	  }
	  *f = type;
	  return;
	}
      }
    }

    /* chase unqual ptrs */
    prevunql = *f;
    unql = g->nodeBuffer[*f].type.unql;
    while(unql != prevunql){
      if (unql == -1){
	fprintf(stderr,"unql chain for %d leads to -1\n",*f);
	exit(1);
      }
      prevunql = unql;
      unql = g->nodeBuffer[prevunql].type.unql;
    }

    /* prevunql and unql point at same node */
    /* end of unql chain */
    if (unql != *f) {
      *f = unql;
    }

  }
}

void do_aggregate(gd_graph *g,int i){
  /* fields to save */
  /* from type_decl */
  node_ptr decl;	/* the type_decl associated with the record type	*/
  node_ptr name;	/*   |- the gcc identifier node if it exists 		*/
  str_ptr srcf;		/*   |- source file					*/
  int srcl;		/*   |- source line					*/
  str_ptr namestr;	/*   \- str_ptr for name				*/

  /* from record_type or union_type */
  node_ptr fields;	/* the first field of the record or union */
  node_ptr functs;	/* the first function of the record or union */
  node_ptr binfo;	/* binfo node -- used for class inheritance */
  t_aggr_enum newaggr;	/* datrix type of aggregation */

  /* other locals */
  node_ptr new_node, prev_node, tmp;

  /* name srcf and srl from the decl */
  decl = g->nodeBuffer[i].type.name;
  if (decl != -1){
    srcf = g->nodeBuffer[decl].decl.srcf;
    srcl = g->nodeBuffer[decl].decl.srcl;
    name = g->nodeBuffer[decl].decl.name;
    if (name == -1){
      namestr = -1;
    } else {
      namestr = g->nodeBuffer[name].identifier_node.strg;
    }
  } else {
    srcf = -1;
    srcl = -1;
    name = -1;
    namestr = -1;
  }

  /* save fields of interest */
  if (g->nodeBuffer[i].common.type == gRECORD_TYPE) {
    fields = g->nodeBuffer[i].record_type.flds;
    functs = g->nodeBuffer[i].record_type.fncs;
    binfo  = g->nodeBuffer[i].record_type.binf;
    if (g->nodeBuffer[i].record_type.isClass){
      newaggr = aggr_class;
    } else {
      newaggr = aggr_struct;
    }
  }
  if (g->nodeBuffer[i].common.type == gUNION_TYPE) {
    fields = g->nodeBuffer[i].union_type.flds;
    functs = g->nodeBuffer[i].union_type.fncs;
    binfo  = g->nodeBuffer[i].union_type.binf;
    newaggr = aggr_union;
  }
  if (binfo ==  0) {
    printf("===== binfo = 0, i = %d\n",i);
  }

  /* switch node type */
  g-> nodeBuffer[i].common.type = dC_AGGR_TYPE;
  g-> nodeBuffer[i].common.other_node = -1;

  g-> nodeBuffer[i].dc_asg_nd.file = srcf;
  g-> nodeBuffer[i].dc_asg_nd.line = srcl;

  g-> nodeBuffer[i].dc_identifier.visb = visb_unk;
  if (decl != -1){
    if (g->nodeBuffer[decl].common.private_flag){
      g->nodeBuffer[i].dc_identifier.visb = visb_pri;
    }
    if (g->nodeBuffer[decl].common.public_flag){
      g->nodeBuffer[i].dc_identifier.visb = visb_pub;
    }
    if (g->nodeBuffer[decl].common.protected_flag){
      g->nodeBuffer[i].dc_identifier.visb = visb_pro;
    }
  } else {
    if (g->nodeBuffer[i].common.private_flag){
      g->nodeBuffer[i].dc_identifier.visb = visb_pri;
    }
    if (g->nodeBuffer[i].common.public_flag){
      g->nodeBuffer[i].dc_identifier.visb = visb_pub;
    }
    if (g->nodeBuffer[i].common.protected_flag){
      g->nodeBuffer[i].dc_identifier.visb = visb_pro;
    }
  }
  g->nodeBuffer[i].dc_identifier.name = namestr;
  g->nodeBuffer[i].dc_identifier.instance = -1;
  g->nodeBuffer[i].dc_identifier.defLoc = -1;
  g->nodeBuffer[i].dc_identifier.arcFriend = -1;
  g->nodeBuffer[i].dc_identifier.array_dim = -1;
  g->nodeBuffer[i].dc_type.implements = -1;
  g-> nodeBuffer[i].dc_aggr_type.aggr_type = newaggr;

  /* do inheritance later */
  g-> nodeBuffer[i].dc_aggr_type.inheritance = binfo;

  /* add members */
  g->nodeBuffer[i].dc_aggr_type.members = -1;
  while(fields != -1){
    /* special case, one of the fields is a type_decl pointing back to us!! */
    if (!(g-> nodeBuffer[fields].common.type == gTYPE_DECL &&
	  g-> nodeBuffer[fields].decl.type == i)){
      new_node = newNode(g);
      g-> nodeBuffer[new_node].common.type = dC_ARC_SON;
      g-> nodeBuffer[new_node].common.other_node = -1;
      g-> nodeBuffer[new_node].dc_arc.src = i;
      g-> nodeBuffer[new_node].dc_arc.dest = fields;
      g-> nodeBuffer[new_node].dc_arc.ordinal = 0;
      g-> nodeBuffer[new_node].dc_arc.next = -1;
      if (g->nodeBuffer[i].dc_aggr_type.members == -1){
	g->nodeBuffer[i].dc_aggr_type.members = new_node;
      } else {
	g->nodeBuffer[prev_node].dc_arc.next = new_node;
      }
      prev_node = new_node;
    } 
    tmp = fields;
    fields = g->nodeBuffer[fields].decl.chan;
    g->nodeBuffer[tmp].decl.chan = -1;
  }
  while(functs != -1) {
    /* special case, do not add function_decls that are artificial!! */
    if (!g->nodeBuffer[functs].decl.artificial) {
      if (!(isNamed(g,functs,base_ctor) || isNamed(g,functs,comp_ctor))){
	/* also do not add function_decls whose name is base_ctor or comp_ctor
	 */
	new_node = newNode(g);
	g-> nodeBuffer[new_node].common.type = dC_ARC_SON;
	g-> nodeBuffer[new_node].dc_arc.src = i;
	g-> nodeBuffer[new_node].dc_arc.dest = functs;
	g-> nodeBuffer[new_node].dc_arc.ordinal = 0;
	g-> nodeBuffer[new_node].dc_arc.next = -1;
	if (g->nodeBuffer[i].dc_aggr_type.members == -1){
	  g->nodeBuffer[i].dc_aggr_type.members = new_node;
	} else {
	  g->nodeBuffer[prev_node].dc_arc.next = new_node;
	}
	prev_node = new_node;
      }
    }
    tmp = functs;
    functs = g->nodeBuffer[functs].decl.chan;
    g->nodeBuffer[tmp].decl.chan = -1;
  }

}

int isNamed(gd_graph * g,node_ptr funct_node, str_ptr name) {
  /* only called on function_decls */
  node_ptr name_node;
  name_node = g->nodeBuffer[funct_node].decl.name;
  if (name_node == -1) {
    return 0;
  }
  if (g->nodeBuffer[name_node].common.type != gIDENTIFIER_NODE) {
    return 0;
  }
  if (g-> nodeBuffer[name_node].identifier_node.strg == name){
    return 1;
  }
  return 0;
}
