/* Make various kinds of derived types: arrays, pointers, and references */

/* AJM 2001 Mar 10 */

#include <stdio.h>
#include <gd_graph.h>
#include "gd_utils.h"

// Lookup string in hash table.
int main (int argc, char *argv [])
{
    void do_derived ();
    char *infilename;
    char *outfilename;
    int i;
    gd_graph g;		// the graph

    if (argc != 3)
    {
      fprintf (stderr, "usage: %s ingdfile outgdfile\n", argv [0]);
      exit (1);
    }

    infilename = argv[1];
    outfilename = argv[2];

    g = gd_graph_read_binary_file (infilename, 1.1, 1.0, 1.1);
    if (g. curStringPos == 0) {
	fprintf (stderr,"can't read gd graph on %s\n", infilename);
	exit (1);
    }

    do_derived (&g);

    if (gd_graph_write_binary_file (outfilename, g) == 0)
    {
	fprintf (stderr, "can't write %s\n", outfilename);
	exit (1);
    }

    return 0;
}

// Arrays, references, and pointers have in common that they're derived
// from some base type.   In gcc they are respectively t_array_type,
// t_reference_type, and t_pointer_type.  In Datrix they are
// t_dc_array_type, t_dc_ref_type, and t_dc_ptr_type.

void do_derived (gd_graph *g)
{
    node_ptr ptr;			// Ptr to node which is type

    gd_graph_node new_type;		// New Datrix node for type.
    gd_graph_node new_edge;		// New Datrix node for edge.
    gd_graph_node new_inst;		// New Datrix node for another edge.
    gd_graph_node new_lit;		// New Datrix node for literal.

    // For extracting array dimension.
    node_ptr range_type;		// array range in gcc is a type node
    node_ptr range_max;			// max of range is a constant node
    int array_dimension;
    char literal [30];			// String to contain literal value.

    // The builtin type int.
    str_ptr int_str_ptr = gd_lookup_string (g, "int");
    node_ptr int_type_ptr;

    // Search for "int".
    for (int_type_ptr = 0; int_type_ptr < g-> curNodePos; ++ int_type_ptr)
	if (
		g-> nodeBuffer [int_type_ptr]. common. type == dC_BUILT_IN_TYPE
	    &&	g-> nodeBuffer [int_type_ptr]. dc_identifier. name == int_str_ptr
	    )
	    break;	// found int
	
    if (int_type_ptr == g-> curNodePos)
	int_type_ptr = -1;

    for (ptr = 0; ptr < g-> curNodePos; ++ptr)
	// There are various kinds of derived type.
	if (
		g-> nodeBuffer [ptr]. common. type == gARRAY_TYPE
	    ||	g-> nodeBuffer [ptr]. common. type == gREFERENCE_TYPE
	    ||	g-> nodeBuffer [ptr]. common. type == gPOINTER_TYPE
	)
	{
	    // Make a new type node.

	    switch (g-> nodeBuffer [ptr]. common. type)
	    {
	    case gREFERENCE_TYPE:
		new_type. common. type = dC_REF_TYPE;
		new_type. common. other_node = -1;
		new_type. dc_identifier. array_dim = -1;
		break;
	    case gPOINTER_TYPE:
		new_type. common. type = dC_PTR_TYPE;
		new_type. common. other_node = -1;
		new_type. dc_identifier. array_dim = -1;
		break;
	    case gARRAY_TYPE:
		new_type. common. type = dC_ARRAY_TYPE;
		new_type. common. other_node = -1;
		new_type. dc_identifier. array_dim = newNode (g);

		// Chase down the dimension of the array, and format it.
		range_type = g-> nodeBuffer [ptr]. array_type. domn;
		range_max = g-> nodeBuffer [range_type]. integer_type. max;
		array_dimension = 1 + g-> nodeBuffer [range_max]. integer_cst. low;
		sprintf (literal, "%d", array_dimension);

		// Construct new literal node for the dimension.
		new_lit. common. type = dC_LITERAL;
		new_lit. common. other_node = -1;
		new_lit. dc_expression. refers = -1;
		new_lit. dc_literal. value = gd_add_string (g, literal);
		new_lit. dc_literal. type = newNode (g);	// for new c_instance edge

		// Construct new instance arc for the literal.
		new_inst. common. type = dC_INSTANCE;
		new_inst. common. other_node = -1;
		new_inst. dc_arc. ordinal = 0;
		new_inst. dc_arc. src = newNode (g);		// for the literal
		new_inst. dc_arc. dest = int_type_ptr;
		new_inst. dc_arc. next = -1;

		// Construct new instance arc for the literal and its type (which is int).
		new_edge. common. type = dC_ARC_ARRAY_DIM;
		new_edge. common. other_node = -1;
		new_edge. dc_arc. ordinal = 0;
		new_edge. dc_arc. src = ptr;
		new_edge. dc_arc. dest = new_inst. dc_arc. src;	// same literal
		new_edge. dc_arc. next = -1;

		// Save them.
		g-> nodeBuffer [new_type. dc_identifier. array_dim] = new_edge;
		g-> nodeBuffer [new_edge. dc_arc. dest] = new_lit;
		g-> nodeBuffer [new_lit. dc_literal. type] = new_inst;
		break;
	    }

	    new_type. dc_identifier. name = -1;		// derived, no name.
	    new_type. dc_identifier. visb =
		    g-> nodeBuffer [ptr]. common. public_flag? visb_pub:
		    g-> nodeBuffer [ptr]. common. private_flag? visb_pri:
		    g-> nodeBuffer [ptr]. common. protected_flag? visb_pro:
		    visb_unk
		;
	    new_type. dc_identifier. defLoc = -1;
	    new_type. dc_identifier. arcFriend = -1;
	    new_type. dc_identifier. instance = newNode (g);	// new edge for instance
	    new_type. dc_type. implements = -1;

	    // Make new instance edge.
	    new_edge. common. type = dC_INSTANCE;
	    new_edge. common. other_node = -1;
	    new_edge. dc_arc. ordinal = 0;
	    new_edge. dc_arc. src = ptr;
	    new_edge. dc_arc. next = -1;
	    switch (g-> nodeBuffer [ptr]. common. type)
	    {
	    case gARRAY_TYPE:
		new_edge. dc_arc. dest = g-> nodeBuffer [ptr]. array_type. elts;
		break;
	    case gREFERENCE_TYPE:
		new_edge. dc_arc. dest = g-> nodeBuffer [ptr]. reference_type. refd;
		break;
	    case gPOINTER_TYPE:
		// Pointer to member isn't handled here!
		new_edge. dc_arc. dest = g-> nodeBuffer [ptr]. pointer_type. ptd;
		break;
	    }

	    // save them back
	    g-> nodeBuffer [ptr] = new_type;
	    g-> nodeBuffer [new_type. dc_identifier. instance] = new_edge;
	}
}
