// Make C_OPERATOR nodes from all the different kinds of *exprs in gcc.

/* AJM 2001 Mar 19 */

#include <stdio.h>
#include <gd_graph.h>
#include "gd_utils.h"

int main (int argc, char *argv [])
{
    void do_exprs (gd_graph *);
    void do_expr_fixups (gd_graph *);
    char *infilename;
    char *outfilename;
    int i;
    gd_graph g;		// the graph

    if (argc != 3)
    {
      fprintf (stderr, "usage: %s ingdfile outgdfile\n", argv [0]);
      exit (1);
    }

    infilename = argv[1];
    outfilename = argv[2];

    g = gd_graph_read_binary_file (infilename, 1.1, 1.0, 1.1);
    if (g. curStringPos == 0) {
	fprintf (stderr,"can't read gd graph on %s\n", infilename);
	exit (1);
    }

    do_exprs (&g);
    do_expr_fixups (&g);

    if (gd_graph_write_binary_file (outfilename, g) == 0)
    {
	fprintf (stderr, "can't write %s\n", outfilename);
	exit (1);
    }


    return 0;
}

// The gcc graph allows for all sorts of expressions which are not c++ expressions.
// These are indicated by -1 in the op_type field.

// Some things called _EXPR in gcc are not operator expressions in the Datrix sense
// (example: gINIT_EXPR).  These are omitted from the table below.

// The 'arity' is used as follows.  For convert_x operators, each op of 0 through arity-1
// is used to make an ArcOpd edge.  For skip_x operators, references to the expression
// anywhere in the whole graph are replaced by references to the arity'th subexpression.
// Note that the operator "array" (op0 through op'arity-1) is defined differently for
// each ops_type of expression: for ref_ops, they're immediately after the common block
// (e.g. as in the component_ref structure); for expr_ops, they're immediately after the expr
// block (as in the compound_expr structure); for cast_ops, op0 is the expr.type and the rest
// are as for expr_ops (offset by one, of course).

struct {
    enum gd_node_type node_type;
    enum {convert_x, skip_x, invalid_x} x_type; 	// what to do with it
    enum {ref_ops, expr_ops, cast_ops} ops_type;	// where the ops are
    int arity;						// see note
    op_kind op_type;					// what datrix type should it be (convert only)
} type_to_op [] =
{
    {gCOMPONENT_REF, 		convert_x, ref_ops,  2, dc_field_access},
    {gBIT_FIELD_REF,		invalid_x, ref_ops,  0, -1},	// apparently not used in cpp
    {gINDIRECT_REF,		convert_x, ref_ops,  1, dc_deref}, 
    {gBUFFER_REF,		invalid_x, ref_ops,  0, -1}, 
    {gARRAY_REF, 		convert_x, ref_ops,  2, dc_array_ref},
    {gCOMPOUND_EXPR,		convert_x, expr_ops, 2, dc_comma}, 
    {gMODIFY_EXPR,		convert_x, expr_ops, 2, dc_asgn_eq}, 
    {gTARGET_EXPR,		convert_x, expr_ops, 2, dc_asgn_eq}, 
    {gCOND_EXPR,		convert_x, expr_ops, 3, dc_question},
    {gBIND_EXPR,		invalid_x, expr_ops, 0, -1}, 
    {gCALL_EXPR,		convert_x, expr_ops, 2, dc_fct_call},
    {gMETHOD_CALL_EXPR,		convert_x, expr_ops, 2, dc_fct_call}, 
    {gWITH_CLEANUP_EXPR,	invalid_x, expr_ops, 0, -1}, 
    {gCLEANUP_POINT_EXPR,	invalid_x, expr_ops, 0, -1}, 
    {gPLACEHOLDER_EXPR,		invalid_x, expr_ops, 0, -1},
    {gWITH_RECORD_EXPR,		invalid_x, expr_ops, 0, -1}, 
    {gPLUS_EXPR,		convert_x, expr_ops, 2, dc_bplus}, 
    {gMINUS_EXPR,		convert_x, expr_ops, 2, dc_bminus}, 
    {gMULT_EXPR,		convert_x, expr_ops, 2, dc_mult}, 
    {gTRUNC_DIV_EXPR,		convert_x, expr_ops, 2, dc_div}, 
    {gCEIL_DIV_EXPR,		invalid_x, expr_ops, 0, -1}, 
    {gFLOOR_DIV_EXPR,		invalid_x, expr_ops, 0, -1}, 
    {gROUND_DIV_EXPR,		invalid_x, expr_ops, 0, -1}, 
    {gTRUNC_MOD_EXPR,		convert_x, expr_ops, 2, dc_modulo}, 
    {gCEIL_MOD_EXPR,		invalid_x, expr_ops, 0, -1}, 
    {gFLOOR_MOD_EXPR,		invalid_x, expr_ops, 0, -1}, 
    {gROUND_MOD_EXPR,		invalid_x, expr_ops, 0, -1}, 
    {gRDIV_EXPR,		invalid_x, expr_ops, 0, -1}, 
    {gEXACT_DIV_EXPR,		invalid_x, expr_ops, 0, -1}, 
    {gFIX_TRUNC_EXPR,		invalid_x, expr_ops, 0, -1}, 
    {gFIX_CEIL_EXPR,		invalid_x, expr_ops, 0, -1}, 
    {gFIX_FLOOR_EXPR,		invalid_x, expr_ops, 0, -1}, 
    {gFIX_ROUND_EXPR,		invalid_x, expr_ops, 0, -1}, 
    {gFLOAT_EXPR,		invalid_x, expr_ops, 0, -1}, 
    {gEXPON_EXPR,		invalid_x, expr_ops, 0, -1}, 
    {gNEGATE_EXPR,		convert_x, expr_ops, 1, dc_uminus}, 
    {gMIN_EXPR,			invalid_x, expr_ops, 0, -1}, 
    {gMAX_EXPR,			invalid_x, expr_ops, 0, -1}, 
    {gABS_EXPR,			invalid_x, expr_ops, 0, -1}, 
    {gFFS_EXPR,			invalid_x, expr_ops, 0, -1}, 
    {gLSHIFT_EXPR,		convert_x, expr_ops, 2, dc_lshift}, 
    {gRSHIFT_EXPR,		convert_x, expr_ops, 2, dc_rshift}, 
    {gLROTATE_EXPR,		invalid_x, expr_ops, 0, -1}, 
    {gRROTATE_EXPR,		invalid_x, expr_ops, 0, -1}, 
    {gBIT_IOR_EXPR,		convert_x, expr_ops, 2, dc_bit_or},
    {gBIT_XOR_EXPR,		convert_x, expr_ops, 2, dc_bit_xor},
    {gBIT_AND_EXPR,		convert_x, expr_ops, 2, dc_bit_and},
    {gBIT_ANDTC_EXPR,		invalid_x, expr_ops, 0, -1}, 
    {gBIT_NOT_EXPR,		convert_x, expr_ops, 1, dc_bit_not},
    {gTRUTH_ANDIF_EXPR,		convert_x, expr_ops, 2, dc_and}, 
    {gTRUTH_ORIF_EXPR,		convert_x, expr_ops, 2, dc_or}, 
    {gTRUTH_AND_EXPR,		invalid_x, expr_ops, 0, -1}, 
    {gTRUTH_OR_EXPR,		invalid_x, expr_ops, 0, -1}, 
    {gTRUTH_XOR_EXPR,		invalid_x, expr_ops, 0, -1}, 
    {gTRUTH_NOT_EXPR,		convert_x, expr_ops, 1, dc_not}, 
    {gLT_EXPR,			convert_x, expr_ops, 2, dc_tst_lt}, 
    {gLE_EXPR,			convert_x, expr_ops, 2, dc_tst_le}, 
    {gGT_EXPR,			convert_x, expr_ops, 2, dc_tst_gt}, 
    {gGE_EXPR,			convert_x, expr_ops, 2, dc_tst_ge}, 
    {gEQ_EXPR,			convert_x, expr_ops, 2, dc_tst_eq}, 
    {gNE_EXPR,			convert_x, expr_ops, 2, dc_tst_ne}, 
    {gUNORDERED_EXPR,		invalid_x, expr_ops, 0, -1}, 
    {gORDERED_EXPR,		invalid_x, expr_ops, 0, -1}, 
    {gUNLT_EXPR,		invalid_x, expr_ops, 0, -1}, 
    {gUNLE_EXPR,		invalid_x, expr_ops, 0, -1}, 
    {gUNGT_EXPR,		invalid_x, expr_ops, 0, -1}, 
    {gUNGE_EXPR,		invalid_x, expr_ops, 0, -1}, 
    {gUNEQ_EXPR,		invalid_x, expr_ops, 0, -1}, 
    {gIN_EXPR,			invalid_x, expr_ops, 0, -1}, 
    {gSET_LE_EXPR,		invalid_x, expr_ops, 0, -1}, 
    {gCARD_EXPR,		invalid_x, expr_ops, 0, -1}, 
    {gRANGE_EXPR,		invalid_x, expr_ops, 0, -1}, 
    {gCONVERT_EXPR,		convert_x, cast_ops, 2, dc_static_cast},
    {gNOP_EXPR,			convert_x, cast_ops, 2, dc_static_cast},
    {gNON_LVALUE_EXPR,		skip_x,    expr_ops, 0, -1}, 
    {gSAVE_EXPR,		invalid_x, expr_ops, 0, -1}, 
    {gUNSAVE_EXPR,		invalid_x, expr_ops, 0, -1}, 
    {gRTL_EXPR,			invalid_x, expr_ops, 0, -1}, 
    {gADDR_EXPR,		convert_x, expr_ops, 1, dc_addr}, 
    {gREFERENCE_EXPR,		invalid_x, expr_ops, 0, -1}, 
    {gENTRY_VALUE_EXPR,		invalid_x, expr_ops, 0, -1}, 
    {gCOMPLEX_EXPR,		invalid_x, expr_ops, 0, -1}, 
    {gCONJ_EXPR,		invalid_x, expr_ops, 0, -1}, 
    {gREALPART_EXPR,		invalid_x, expr_ops, 0, -1}, 
    {gIMAGPART_EXPR,		invalid_x, expr_ops, 0, -1}, 
    {gPREDECREMENT_EXPR,	convert_x, expr_ops, 1, dc_pre_dec},  /* TD */
    {gPREINCREMENT_EXPR,	convert_x, expr_ops, 1, dc_pre_inc},  /* TD */
    {gPOSTDECREMENT_EXPR,	convert_x, expr_ops, 1, dc_post_dec}, /* TD */
    {gPOSTINCREMENT_EXPR,	convert_x, expr_ops, 1, dc_post_inc}, /* TD */
    {gVA_ARG_EXPR,		invalid_x, expr_ops, 0, -1}, 
    {gTRY_CATCH_EXPR,		invalid_x, expr_ops, 0, -1}, 
    {gTRY_FINALLY_EXPR,		invalid_x, expr_ops, 0, -1}, 
    {gGOTO_SUBROUTINE_EXPR,	invalid_x, expr_ops, 0, -1}, 
    {gPOPDHC_EXPR,		invalid_x, expr_ops, 0, -1}, 
    {gPOPDCC_EXPR,		invalid_x, expr_ops, 0, -1}, 
    {gLABEL_EXPR,		invalid_x, expr_ops, 0, -1}, 
    {gGOTO_EXPR,		invalid_x, expr_ops, 0, -1}, 
    {gRETURN_EXPR,		invalid_x, expr_ops, 0, -1}, 
    {gEXIT_EXPR,		invalid_x, expr_ops, 0, -1}, 
    {gLOOP_EXPR,		invalid_x, expr_ops, 0, -1}, 
    {gLABELED_BLOCK_EXPR,	invalid_x, expr_ops, 0, -1}, 
    {gEXIT_BLOCK_EXPR,		invalid_x, expr_ops, 0, -1}, 
    {gEXPR_WITH_FILE_LOCATION,	invalid_x, expr_ops, 0, -1}, 
    {gSWITCH_EXPR,		invalid_x, expr_ops, 0, -1}, 
    {gSRCLOC,			invalid_x, expr_ops, 0, -1}, 
    {gSIZEOF_EXPR,		invalid_x, expr_ops, 0, -1}, 
    {gARROW_EXPR,		invalid_x, expr_ops, 0, -1}, 
    {gALIGNOF_EXPR,		invalid_x, expr_ops, 0, -1}, 
    -1
};

void do_exprs (gd_graph *g)
{
    int i;
    int opn;				// Operand number
    node_ptr ptr;			// Ptr to node which is to be converted to "operator".
    gd_graph_node *expr_node_p;		// Pointer to node which is operator

    gd_graph_node new_expr_node;	// New Datrix node for expression ("operator")
    gd_graph_node new_edge;		// New Datrix edge for various purposes.
    node_ptr *ptr_ref;			// Pointer to ptr variable which will pt to the next edge node
    node_ptr ops [3];			// operands

    static void do_ptr (gd_graph *, node_ptr, node_ptr *, int);

    // Do all the expressions.
    for (ptr = 0; ptr < g-> curNodePos; ++ ptr)
    {
	expr_node_p = &(g-> nodeBuffer [ptr]);

	// Consider all the expression types.
	for (i = 0; type_to_op [i]. node_type != -1; ++ i)
	    if (expr_node_p-> common. type == type_to_op [i]. node_type)
	    {
		// Found it!
		switch (type_to_op [i]. x_type)
		{
		case convert_x:
		    // Node conversion from gcc expression to datrix operator
		    new_expr_node. common. type = dC_OPERATOR;
		    new_expr_node. dc_expression. refers = -1;
		    new_expr_node. dc_operator. op = type_to_op [i]. op_type;

		    // Make operand edges.

		    // Get the operands.
		    switch (type_to_op [i]. ops_type)
		    {
			case ref_ops:
			    for (opn = 0; opn < type_to_op [i]. arity; ++opn)
				ops [opn] = (& expr_node_p-> component_ref. op0) [opn];
			    break;
			case expr_ops:
			    for (opn = 0; opn < type_to_op [i]. arity; ++opn)
				ops [opn] = (& expr_node_p-> compound_expr. op0) [opn];
			    break;
			case cast_ops:
			    ops [0] = expr_node_p-> expr. type;
			    for (opn = 1; opn < type_to_op [i]. arity; ++opn)
				ops [1] = expr_node_p-> expr1. op0;
			    break;
		    }

		    // The first ptr is the one in the base expression node.
		    ptr_ref = & new_expr_node. dc_operator. ops;

		    // Walk the ops array.
		    for (opn = 0; opn < type_to_op [i]. arity; ++ opn)
		    {
			// create a Opd edge for ops [opn]
			*ptr_ref = newNode (g);
			g-> nodeBuffer [*ptr_ref]. common. type = dC_ARC_OPD;
			g-> nodeBuffer [*ptr_ref]. common. other_node = -1;
			g-> nodeBuffer [*ptr_ref]. dc_arc. ordinal = opn+1;
			g-> nodeBuffer [*ptr_ref]. dc_arc. src = ptr;
			g-> nodeBuffer [*ptr_ref]. dc_arc. dest = ops [opn];
			g-> nodeBuffer [*ptr_ref]. dc_arc. next = -1;

			// Continue to the next link.
			ptr_ref = & g-> nodeBuffer [*ptr_ref]. dc_arc. next;
		    }
		    *expr_node_p = new_expr_node;
		    break;

		case skip_x:
		    // Node removal.  For uniformity, make a placeholder (gcc) expression
		    // with other_node pointing at the skip-to.   Fix up references later.
		    expr_node_p-> common. type = gPLACEHOLDER_EXPR;

		    // Pick out the right operand of the skippable operator, and drop it
		    // into the other_node ptr.
		    switch (type_to_op [i]. ops_type)
		    {
			case ref_ops:
			    expr_node_p-> common. other_node =
				(& expr_node_p-> component_ref. op0) [type_to_op [i]. arity];
			    break;
			case expr_ops:
			    expr_node_p-> common. other_node =
				(& expr_node_p-> compound_expr. op0) [type_to_op [i]. arity];
			    break;
			case cast_ops:
			    expr_node_p-> common. other_node =
				(type_to_op [i]. arity == 0?
				    expr_node_p-> expr. type:
				    (& expr_node_p-> expr1. op0) [type_to_op [i]. arity - 1]
				);
			    break;
		    }
		    break;

		case invalid_x:
		    // Unknown expression.
		    fprintf (stderr, "gd_operators: node %d of type %d is an unexpected expression.\n",
			    ptr,
			    expr_node_p-> common. type
			);
		    // Erase the other node so we don't try to skip it.
		    expr_node_p-> common. other_node = -1;
		    break;

		default:
		    fprintf (stderr, "gd_operators: impossible\n");
		    exit (1);
		}
		break; // found it, done.
	    }
    }

    // Now do everything everywhere, replacing ptrs to placeholder_exprs by their other_nodes.
    walk_ptrs (g, do_ptr);
}

// What to do at each node when walking.
// What to do is: p is a ptr to a placeholder_expr node, skip it to its other_node.
static void do_ptr (gd_graph *g, node_ptr np, node_ptr *p, int n)
{
    gd_graph_node *pp;

    // Do nothing if *p is not a valid ptr (including if it's null).
    if (*p < 0 || *p >= g-> curNodePos)
	return;

    // Do nothing if this is an other_node ptr.
    if (n == 0)
	return;

    pp = & g-> nodeBuffer [*p];

    if (
	    pp-> common. type == gPLACEHOLDER_EXPR
	&&  pp-> common. other_node >= 0
	&&  pp-> common. other_node < g-> curNodePos)

	// The node_ptr variable *p pts at a skippable node.
	// Skip that node, by making *p pt at the node it should skip to.
	// A ptr to that was saved as the other_node ptr of the node *p pts at now.
	*p = pp-> common. other_node;
}

void do_expr_fixups (gd_graph *g){
  /* several fixups have to be made to the expression graphs:
   * 1) Reference to function names have an addr operator in the graph
   *        funct_call -> ArcOpd(1) -> addr -> ArcOpd -> Function
   *     => funct_call -> ArcOpd(1) -> function
   *        asgn_eq -> ArcOpd(2) -> addr -> ArcOpd -> Function
   *     => asgn_eq -> ArcOpd(2) -> Function
   * 2) References to string literals have
   */

  node_ptr i;

  for(i = 0; i < g->curNodePos; i++){
    if (g->nodeBuffer[i].common.type == dC_ARC_OPD){
      node_ptr dest1 = g->nodeBuffer[i].dc_arc.dest;
      if (g->nodeBuffer[dest1].common.type == dC_OPERATOR){
	if (g-> nodeBuffer[dest1].dc_operator.op == dc_addr){
	  /* address has one opd */
	  node_ptr addr_opd, addr_op;
	  addr_opd = g->nodeBuffer[dest1].dc_operator.ops;
	  addr_op = g->nodeBuffer[addr_opd].dc_arc.dest;
	  /* fprintf(stderr,"Opd1 (%d)\n",i);
	  fprintf(stderr,"Addr (%d)\n",dest1);
	  fprintf(stderr,"ops: %d\n",addr_opd);
	  fprintf(stderr,"tgt: %d\n",addr_op); */
	  if (g->nodeBuffer[addr_op].common.type == dC_FUNCTION){
	    /* (case1) address of a function  delete addr op*/
	    /* start by pointing first ArcOpd to the function directly */
	    g->nodeBuffer[i].dc_arc.dest = addr_op;
	    /* delete addr op node and the opd that pointed at the nameref */
	    g->nodeBuffer[dest1].common.type = NODE_DELTED;
	    g->nodeBuffer[addr_opd].common.type = NODE_DELTED;
	  }
	}
      }
    }
  }
}
